#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'

from datetime import datetime

from elasticsearch_dsl import Q, Search
from pyramid.threadlocal import get_current_registry
from pyramid_es import get_client

from pyams_catalog.query import CatalogResultSet
from pyams_content.features.search import ISearchFolder
from pyams_content.shared.common import IGNORED_CONTENT_TYPES
from pyams_content.shared.view.interfaces import IViewQuery, \
    IViewQueryEsParamsExtension, IViewQueryFilterExtension, IViewUserQuery, IWfView, RELEVANCE_ORDER
from pyams_content.shared.view.portlet import SEARCH_EXCLUDED_ITEMS
from pyams_sequence.interfaces import ISequentialIntIds
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.list import unique_iter
from pyams_utils.registry import get_utility
from pyams_utils.timezone import tztime
from pyams_workflow.interfaces import IWorkflow


@adapter_config(name='es', context=IWfView, provides=IViewQuery)
class EsViewQuery(ContextAdapter):
    """View query for Elasticsearch"""

    def get_params(self, context, request=None):
        view = self.context
        registry = get_current_registry()
        # check publication dates
        now = tztime(datetime.utcnow())
        params = Q('range', **{'workflow.effective_date': {'lte': now}})
        # check workflow states
        wf_params = []
        for workflow in registry.getAllUtilitiesRegisteredFor(IWorkflow):
            wf_params.extend(workflow.visible_states)
        params &= Q('terms', **{'workflow.status': wf_params})
        # check custom extensions
        for name, adapter in sorted(registry.getAdapters((view,), IViewQueryEsParamsExtension),
                                    key=lambda x: x[1].weight):
            for new_params in adapter.get_es_params(context, request):
                if new_params is None:
                    return None
                else:
                    params &= new_params
        # activate search
        params &= Q('bool', must=Q('range', **{'workflow.push_end_date': {'gte': now}})) | \
                  Q('bool', must_not=Q('exists', **{'field': 'workflow.push_end_date'}))
        # check content path
        content_path = view.get_content_path(context)
        if content_path is not None:
            params &= Q('term', **{'parent_ids': content_path})
        # check content types
        params &= Q('bool', must_not=Q('terms', **{'content_type': tuple(IGNORED_CONTENT_TYPES.keys())}))
        content_types = view.get_content_types(context)
        if content_types:
            params &= Q('terms', **{'content_type': content_types})
        # check data types
        data_types = view.get_data_types(context)
        if data_types:
            params &= Q('terms', **{'data_type': data_types})
        return params

    def get_results(self, context, sort_index, reverse, limit, request=None):
        registry = request.registry
        client = get_client(request)
        params = self.get_params(context, request)
        if params is None:
            items = CatalogResultSet([])
            total_count = 0
        else:
            search = Search(using=client.es, index=client.index) \
                .params(request_timeout=30) \
                .query(params) \
                .source(['internal_id'])
            # Define sort order
            sort_values = []
            if (not sort_index) or (sort_index == RELEVANCE_ORDER):
                sort_values.append({
                    '_score': {
                        'order': 'desc'
                    }
                })
            else:
                sort_values.append({
                    'workflow.{0}'.format(sort_index): {
                        'order': 'desc' if reverse else 'asc',
                        'unmapped_type': 'date'
                    }
                })
            if sort_values:
                search = search.sort(*sort_values)
            # Define limits
            if limit:
                search = search[:limit]
            else:
                search = search[:999]
            # Get query results
            items = CatalogResultSet([result.internal_id for result in search])
            total_count = search.count()
        for name, adapter in sorted(registry.getAdapters((self.context,), IViewQueryFilterExtension),
                                    key=lambda x: x[1].weight):
            items = adapter.filter(context, items, request)
        return unique_iter(items), total_count


#
# Custom search folders extensions
#

@adapter_config(name='es', context=ISearchFolder, provides=IViewQuery)
class EsSearchFolderQuery(EsViewQuery):
    """Search folder view query for Elasticsearch"""

    def get_params(self, context, request=None):
        params = super(EsSearchFolderQuery, self).get_params(context, request)
        registry = request.registry
        for name, adapter in registry.getAdapters((self,), IViewUserQuery):
            for user_param in adapter.get_user_params(request):
                params &= user_param
        return params


@adapter_config(name='exclusions', context=EsSearchFolderQuery, provides=IViewUserQuery)
class EsSearchFolderExclusionsQuery(ContextAdapter):
    """Search folder exclusions for Elasticsearch

    This adapter is looking into request's annotations for items which should be excluded
    from search.
    """

    @staticmethod
    def get_user_params(request):
        excluded_items = request.annotations.get(SEARCH_EXCLUDED_ITEMS)
        if excluded_items:
            yield Q('bool', must_not=Q('terms', **{'reference_id': tuple(excluded_items)}))


@adapter_config(name='user_search', context=EsSearchFolderQuery, provides=IViewUserQuery)
class EsSearchFolderUserQuery(ContextAdapter):
    """Search folder user query for Elasticsearch

    This adapter is looking for any fulltext search entered by user
    """

    @staticmethod
    def get_user_params(request):
        params = request.params
        fulltext = params.get('user_search')
        if fulltext:
            if fulltext.startswith('+'):
                sequence = get_utility(ISequentialIntIds)
                oid = sequence.get_full_oid(fulltext)
                yield Q('term', **{'reference_id': oid})
            else:
                yield Q('simple_query_string', **{'query': fulltext,
                                                  'fields': [
                                                      'title.*^5.0',
                                                      'header.*^3.0'
                                                  ]}) | \
                      Q('simple_query_string', **{'query': fulltext})
